import re

from dante_tokenizer.data.load import read_test_data


def longest_common_token_sequence(left: list, right: list) -> int:
    """
    Implementation of Longest Common Subsequence using lists of tokens
    instead of strings, hence Longest Common Token Sequence.

    Parameters
    ----------
    left: list
        First first of tokens
    right: list
        Second list of tokens
    i_left: int
        Index for iterating over left list
    i_right: int
        Index for iterating over right list

    Returns
    -------
    int
        Size of the Logest Common Token Sequence found.
    """

    n_left = len(left)
    n_right = len(right)

    L = [[None]*(n_right+1) for i in range(n_left+1)]

    for i in range(n_left+1):
        for j in range(n_right+1):
            if i == 0 or j == 0 :
                L[i][j] = 0
            elif left[i-1] == right[j-1]:
                L[i][j] = L[i-1][j-1]+1
            else:
                L[i][j] = max(L[i-1][j] , L[i][j-1])

    return L[n_left][n_right]

def evaluate_dataset(pred_tokens: list, true_tokens: list, complete_metrics:bool = False) -> (float, float):
    """
    Evaluate tokenizer's results, returning precision, recall and extra_metrics if
    `complete_metrics` is set to True. 

    Parameters
    ----------
    pred_tokens: list
        List of predicted tokens.
    true_tokens: list
        List of true tokens.
    complete_metrics: bool
        If true, return aditional metrics.
    
    Returns
    -------
    (float, float)
        Precision and Recall measurements.
    """
    assert len(pred_tokens) == len(true_tokens)
    n_sents = len(pred_tokens)
    true_positives = 0
    false_positives = 0
    false_negatives = 0
    incorrect_sentences_ids = []

    for i in range(n_sents):
        tp, fp, fn = evaluate_sentence(pred_tokens[i], true_tokens[i])
        true_positives += tp
        false_positives += fp
        false_negatives += fn

        if fp > 0 or fn > 0:
            incorrect_sentences_ids.append(i)

    precision = true_positives / (true_positives + false_positives + 1e-9)
    recall = true_positives / (true_positives + false_negatives + 1e-9)

    if complete_metrics:
        additional_metrics = {
            "true_positives": true_positives,
            "false_positives": false_positives,
            "false_negatives": false_negatives,
            "incorrect_sentences_ids": incorrect_sentences_ids
        }
        return precision, recall, additional_metrics

    return precision, recall

def evaluate_sentence(pred_tokens: list, true_tokens: list) -> (int, int):
    """
    Calculates precision and coverage of tokens.
    Based on the Longest Common Subsequence Algorithm
    to evaluate pred_tokens and true_tokens, with the difference of
    not considering spaced substrings as matches.

    Examples
    --------
    >>> evaluate_sentence(["Oi", ":)"], ["Oi", ":)"])
    (2, 0, 0)

    Parameters
    ----------
    pred_tokens: list
        List of tokens generated by the tokenizer
    true_tokens: list
        List of true tokens

    Returns
    -------
    (int, int, int)
        True positive, false positives, false negatives
    """

    true_positives = longest_common_token_sequence(pred_tokens, true_tokens)
    false_positives = len(pred_tokens) - true_positives
    false_negatives = len(true_tokens) - true_positives

    return true_positives, false_positives, false_negatives



from dante_tokenizer.evaluate import evaluate_sentence


def test_evaluate_sentence():
    """ Tests evaluate_sentence function on documented cases for DANTE dataset """

    # Case 1: Stock names
    assert evaluate_sentence(["Comprei", "PETR4", ","], ["Comprei", "PETR4", ","]) == (3, 0, 0)
    assert evaluate_sentence(["Comprei", "PETR", "4", ","], ["Comprei", "PETR4", ","]) == (2, 2, 1)

    # Case 2: "para" and "com" abbreviation
    assert evaluate_sentence(["Estou", "c/", "MGLU3"],
                             ["Estou", "c/", "MGLU3"]) == (3, 0, 0)
    assert evaluate_sentence(["Estou", "c", "/", "MGLU3"],
                             ["Estou", "c/", "MGLU3"]) == (2, 2, 1)
    assert evaluate_sentence(["Estou", "p/", "MGLU3"],
                             ["Estou", "p/", "MGLU3"]) == (3, 0, 0)
    assert evaluate_sentence(["Estou", "p", "/", "MGLU3"],
                             ["Estou", "p/", "MGLU3"]) == (2, 2, 1)

    # Case 3: Cash symbol
    assert evaluate_sentence(["Tenho", "R$", "1.000,00"],
                             ["Tenho", "R$", "1.000,00"]) == (3, 0, 0)
    assert evaluate_sentence(["Tenho", "R", "$", "1.000,00"],
                             ["Tenho", "R$", "1.000,00"]) == (2, 2, 1)
    assert evaluate_sentence(["Tenho", "U$", "1.000,00"],
                             ["Tenho", "U$", "1.000,00"]) == (3, 0, 0)
    assert evaluate_sentence(["Tenho", "U", "$", "1.000,00"],
                             ["Tenho", "U$", "1.000,00"]) == (2, 2, 1)

    # Case 4: Emotions and Emojis
    assert evaluate_sentence(["Oi", ":)"], ["Oi", ":)"]) == (2, 0, 0)
    assert evaluate_sentence(["Oi", ":", ")"], ["Oi", ":)"]) == (1, 2, 1)
    assert evaluate_sentence(["Oi", ":("], ["Oi", ":("]) == (2, 0, 0)
    assert evaluate_sentence(["Oi", ":", "("], ["Oi", ":("]) == (1, 2, 1)
    assert evaluate_sentence(["Oi", "=)"], ["Oi", "=)"]) == (2, 0, 0)
    assert evaluate_sentence(["Oi", "=", ")"], ["Oi", "=)"]) == (1, 2, 1)
    assert evaluate_sentence(["Oi", "=("], ["Oi", "=("]) == (2, 0, 0)
    assert evaluate_sentence(["Oi", "=", "("], ["Oi", "=("]) == (1, 2, 1)

    # Case 5: Data formats
    assert evaluate_sentence(["07/01/1997"], ["07/01/1997"]) == (1, 0, 0)
    assert evaluate_sentence(["07/01"], ["07/01"]) == (1, 0, 0)
    assert evaluate_sentence(["1997/01/07"], ["1997/01/07"]) == (1, 0, 0)
    assert evaluate_sentence(["07", "/", "01", "/", "1997"], ["07/01/1997"]) == (0, 5, 1)
    assert evaluate_sentence(["07", "/", "01"], ["07/01"]) == (0, 3, 1)
    assert evaluate_sentence(["1997", "/", "01", "/", "07"], ["1997/01/07"]) == (0, 5, 1)

    # Case 6: Mention and hashtags
    assert evaluate_sentence(["@elonmusk", "rocks"],
                             ["@elonmusk", "rocks"]) == (2, 0, 0)
    assert evaluate_sentence(["@", "elonmusk", "rocks"],
                             ["@elonmusk", "rocks"]) == (1, 2, 1)
    assert evaluate_sentence(["Lex", "Fridman", "#robot"],
                             ["Lex", "Fridman", "#robot"]) == (3, 0, 0)
    assert evaluate_sentence(["Lex", "Fridman", "#", "robot"],
                             ["Lex", "Fridman", "#robot"]) == (2, 2, 1)

    # Case 7: Stock market terminology
    assert evaluate_sentence(["$CTIP3", "rendeu"], ["$CTIP3", "rendeu"]) == (2, 0, 0)
    assert evaluate_sentence(["$", "CTIP3", "rendeu"], ["$CTIP3", "rendeu"]) == (1, 2, 1)
    assert evaluate_sentence(["alta", "no", "estocástico", "%k"],
                             ["alta", "no", "estocástico", "%k"]) == (4, 0, 0)
    assert evaluate_sentence(["alta", "no", "estocástico", "%", "k"],
                             ["alta", "no", "estocástico", "%k"]) == (3, 2, 1)



import argparse

from dante_tokenizer.data.load import read_test_data
from dante_tokenizer.evaluate import evaluate_dataset
from dante_tokenizer.tokenizer import (
    predict_dante_tokenizer, 
    predict_nltk_twitter_tokenizer, 
    predict_nltk_word_tokenizer
)


def main():

    parser = argparse.ArgumentParser("Evaluate different tokenizers on Dante Dataset (Brazilian Stock-Market Tweets)")
    parser.add_argument("csv_path", type=str, help="Path to the Dante dataset csv file")
    parser.add_argument("conllu_path", type=str, help="Path to the conllu file containing tokenized sentences")
    parser.add_argument("--debug", default=False, action="store_true", 
                        help="Print detailed metrics and wrong sentence tokens")
    args = parser.parse_args()

    ids, sentences, true_tokens = read_test_data(args.csv_path, args.conllu_path)
    
    tokenizers = [
        # ("nltk Word Tokenizer", predict_nltk_word_tokenizer),
        # ("nltk Twitter Tokenizer", predict_nltk_twitter_tokenizer),
        ("DANTE Tokenizer", predict_dante_tokenizer)
    ]

    for name, tokenizer in tokenizers:
        pred_tokens = tokenizer(sentences)

        precision, recall, extra_metrics = evaluate_dataset(pred_tokens, true_tokens, complete_metrics=True)
        
        print((f"{name} precision: {precision*100}, recall: {recall*100} true_positives: {extra_metrics['true_positives']} "
               f"false_positives: {extra_metrics['false_positives']} false_negatives: {extra_metrics['false_negatives']} "))

        if args.debug:
            for incorrect_sentences_id in extra_metrics["incorrect_sentences_ids"]:
                print("*"*20)
                print(pred_tokens[incorrect_sentences_id])
                print(true_tokens[incorrect_sentences_id])
            print()

if __name__ == "__main__":
    main()


import re

from dante_tokenizer.data.preprocessing import expand_contractions, remove_quotes

def read_tokens_from_csv(csv_path: str) -> list:
    """
    """
    sent_ids = []
    sent_texts = []

    csv_split_regex = r"(?:,|\n|^)(\"(?:(?:\"\")*[^\"]*)*\"|[^\",\n]*|(?:\n|$))"

    csv_file = open(csv_path, "r")
    csv_data = csv_file.readlines()

    for csv_line in csv_data[1:]:

        tokens_matches = re.findall(csv_split_regex, csv_line)

        sent_id = tokens_matches[0]
        sent_text = tokens_matches[1]
        sent_text = remove_quotes(sent_text)
        sent_text = expand_contractions(sent_text)

        sent_ids.append(sent_id)
        sent_texts.append(sent_text)

    csv_file.close()

    return sent_ids, sent_texts

def read_test_data(csv_path: str, conllu_path: str) -> dict:
    """
    Read the unparsed sentences from a csv formatted file, expecting the second 
    column to contain the raw sentence. Secondily, reads the formatted dataset 
    on the CoNNL-U format to extract sentence's tokens.

    Obs: This function assumes that the row number of the csv_path file corresponds
    to the conllu_path file `sent_id` field.

    Parameters
    ----------
    conllu_path: str
        Path to the conllu formatted dataset.

    Returns
    -------
    list
        List containing sentence ids.
    list
        List containing sentence original texts.
    list
        List containing a list of sentence tokens.
    """

    sent_ids = []
    sent_texts = []
    sent_tokens = []

    # TODO: Fix this regex, Won't work with last line from conllu file.
    conllu_sentence_regex = r"(# newdoc[\s\S]*?[\r\n]{2})"
    conllu_sentence_id_regex = r"sent_id = ([\d]+)"
    conllu_token_split_regex = r"\d\t([^\t]*)"
    csv_split_regex = r"(?:,|\n|^)(\"(?:(?:\"\")*[^\"]*)*\"|[^\",\n]*|(?:\n|$))"

    csv_file = open(csv_path, "r")
    csv_data = csv_file.readlines()
    conllu_file = open(conllu_path, "r")
    conllu_data = conllu_file.read()

    try:
        matches = re.finditer(conllu_sentence_regex, conllu_data)

        for matchNum, match in enumerate(matches, start=1):
            conllu_text = match.group(0)
            id_matches = re.finditer(conllu_sentence_id_regex, conllu_text)
            for id_match in id_matches:
                sent_id = int(id_match.group(1)) 
            tokens = []
            
            token_matches = re.finditer(conllu_token_split_regex, conllu_text)
            for token_match in token_matches:
                tokens.append(token_match.group(1))

            csv_line = csv_data[sent_id]
            csv_matches = re.findall(csv_split_regex, csv_line)
            sentence = csv_matches[1]
            # Pre-processing text
            sentence = remove_quotes(sentence)
            sentence = expand_contractions(sentence)

            sent_ids.append(sent_id)
            sent_texts.append(sentence)
            sent_tokens.append(tokens)

    finally:

        csv_file.close()
        conllu_file.close()

    return sent_ids, sent_texts, sent_tokens


from setuptools import setup, find_packages

setup(
    name="dante-tokenizer",
    version="0.1",
    author="Emanuel Huber",
    author_email="emanuel.huber@usp.br",
    packages=find_packages(),
    license="Apache 2.0",
    description="""""",
    url="https://github.com/huberemanuel/twitter-portuguese-tokenizer",
    download_url="https://github.com/huberemanuel/twitter-portuguese-tokenizer/archive/refs/tags/v0.1-alpha.tar.gz",
    keywords=["tokenizer", "twitter", "portuguese"],
    install_requires=[
        "nltk>=3.6.2",
        "regex>=2021.4.4"
    ],
    classifier=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers|Researchers",
        "Topic :: Natural Language Processing :: Tokenizer",
        "License :: OSI Approved :: Apache 2.0",
        "Programming Language :: Python :: 3.8"
    ]
)


import argparse
import os

from dante_tokenizer.tokenizer import DanteTokenizer
from dante_tokenizer.data.load import read_tokens_from_csv


def tokenize_csv(csv_path: str) -> list:
    """ 
    Runs DanteTokenizer for each row in the csv file, it is expected
    to the first column be the sentence id and the second to be the
    sentence text.

    Parameters
    ----------
    csv_path: str
        Full path to the csv input file.

    Returns
    -------
    list
        List of parsed tokens for each sentence.
    """
    
    sent_ids, sent_texts = read_tokens_from_csv(csv_path)
    tokenizer = DanteTokenizer()
    sent_tokens = list(map(tokenizer.tokenize, sent_texts))

    return sent_ids, sent_tokens, sent_texts

def tokens_to_conllu(doc_name: str, sent_ids: list, sent_tokens: list, 
                     sent_texts: list) -> list:

    """
    Create conllu string representation based on original sentences and tokens.

    Parameters
    ----------
    doc_name: str
        Name of the given doc.
    sent_ids: list
        List of sentence ids.
    sent_tokens: list
        List of parsed tokens.
    sent_texts: list
        List of sentence texts.

    Returns:
        List of strings on the conllu format.
    """

    conllu_text = [""]

    for sent_id, sent_token, sent_text in zip(sent_ids, sent_tokens, sent_texts):
        conllu_text.append(f"# newdoc id = {doc_name}\n")
        conllu_text.append("# newpar\n")
        conllu_text.append(f"# sent_id = {sent_id}\n")
        conllu_text.append(f"# text = {sent_text}\n")

        for token_id, token in enumerate(sent_token, start=1):
            conllu_text.append(f"{token_id}\t{token}" + "\t_" * 7)
            if token_id == len(sent_token):
                conllu_text[-1] += "\tSpacesAfter=\\n\n"
            else:
                conllu_text[-1] += "\t_\n"
        conllu_text.append("\n")

    return conllu_text

def save_conllu(conllu_text: list, file_name: str) -> None:
    """ 
    Creates conllu file.
    
    Parameters
    ----------
    conllu_text: list
        List of strings on the conllu format
    file_name: str
        Output path to store the conllu file.
    """
    conllu_file = open(file_name, "w")

    for line in conllu_text:
        conllu_file.writelines(line)

    conllu_file.close()

def main():
    parser = argparse.ArgumentParser("Transforms csv file to conllu file")
    parser.add_argument("csv_path", type=str, help="Full path to the csv file")
    args = parser.parse_args()
    doc_name = os.path.basename(args.csv_path)
    sent_ids, sent_tokens, sent_texts = tokenize_csv(args.csv_path)
    conllu_text = tokens_to_conllu(doc_name, sent_ids, sent_tokens, sent_texts)
    save_conllu(conllu_text, f"{doc_name}.conllu")

if __name__ == "__main__":
    main()